/**************************************************************************
 *  Copyright (C) 2008 by Simon Qian                                      *
 *  SimonQian@SimonQian.com                                               *
 *                                                                        *
 *  Project:    vsprog                                                    *
 *  File:       versaloon.c                                               *
 *  Author:     SimonQian                                                 *
 *  Versaion:   See changelog                                             *
 *  Purpose:    common functions for Versaloon                            *
 *  License:    See license                                               *
 *------------------------------------------------------------------------*
 *  Change Log:                                                           *
 *      YYYY-MM-DD:     What(by Who)                                      *
 *      2008-10-29:     created(by SimonQian)                             *
 **************************************************************************/
#include "app_cfg.h"
#include "usb.h"

#include "versaloon.h"
#include "usbtoxxx/usbtoxxx.h"


const char *versaloon_hardwares[] = {"Versaloon_Full", "Versaloon_Mini", "Versaloon_Nano",};
int versaloon_to = VERSALOON_TIMEOUT;

uint8 *versaloon_buf = NULL;

usb_dev_handle *versaloon_device_handle = NULL;
uint16 versaloon_vid = VERSALOON_VID, versaloon_pid = VERSALOON_PID;
uint8 versaloon_epout = VERSALOON_OUTP, versaloon_epin = VERSALOON_INP;
uint16 versaloon_buf_size = 256;

versaloon_pending_result_t versaloon_pending_result[VERSALOON_MAX_PENDING_NUMBER];
uint16 versaloon_pending_idx = 0;


void versaloon_usage(void)
{
	printf("usage of %s:\n"
			"vsprog [COMMON_ARGUMENTS]\n"
			"  <-p vid pid epin epout> usb device and Port, default is Versaloon(0x03EB 0x2103 0x82 0x02)\n"
		, VERSALOON_STRING);
}

void versaloon_support(void)
{
	printf("%s: see http://www.SimonQian.com/en/Versaloon\n", VERSALOON_STRING);
}

RESULT versaloon_check_argument(char cmd, const char **argu, uint32 argu_num)
{
	switch(cmd)
	{
	case 'h':
		versaloon_usage();
		break;
	case 'S':
		versaloon_support();
		break;
	case 'P':
		if(argu_num != 4) { LOG_ERROR("args of %c error\n", cmd); return ERROR_FAIL; }
		versaloon_set_usb_param((uint16)strtoul(argu[0], NULL, 0),	// vid
								(uint16)strtoul(argu[1], NULL, 0),	// pid
								(uint8)strtoul(argu[2], NULL, 0),	// epin
								(uint8)strtoul(argu[3], NULL, 0));	// epout
		break;
	default:
		return ERROR_FAIL;
		break;
	}

	return ERROR_OK;
}



void versaloon_set_usb_param(uint16 vid, uint16 pid, uint8 epin, uint8 epout)
{
	versaloon_vid	= vid;
	versaloon_pid	= pid;
	versaloon_epin	= epin;
	versaloon_epout	= epout;
}

RESULT versaloon_add_pending(uint8 type, uint8 cmd, uint16 actual_szie, uint16 want_pos, uint16 want_size, uint8 *buffer, uint8 collect)
{
	if(versaloon_pending_idx >= VERSALOON_MAX_PENDING_NUMBER) { LOG_ERROR("not enough pending struct\n"); return ERROR_FAIL; }

	versaloon_pending_result[versaloon_pending_idx].type = type;
	versaloon_pending_result[versaloon_pending_idx].cmd = cmd;
	versaloon_pending_result[versaloon_pending_idx].actual_data_size = actual_szie;
	versaloon_pending_result[versaloon_pending_idx].want_data_pos = want_pos;
	versaloon_pending_result[versaloon_pending_idx].want_data_size = want_size;
	versaloon_pending_result[versaloon_pending_idx].data_buffer = buffer;
	versaloon_pending_result[versaloon_pending_idx].collect = collect;
	versaloon_pending_idx++;

	return ERROR_OK;
}

RESULT versaloon_send_command(uint16 out_len, uint16 *inlen)
{
	int ret;

#ifdef PARAM_CHECK
	if(NULL == versaloon_buf) { LOG_BUG("buffer not valid\n"); return ERROR_FAIL; }
	if(0 == out_len) { LOG_BUG("parameter invalid\n"); return ERROR_OK; }
#endif

	ret = usb_bulk_write(versaloon_device_handle, versaloon_epout, (char *)versaloon_buf, out_len, versaloon_to);
	if(ret != out_len) { LOG_ERROR("usb send failed with %d\n", ret); return ERROR_FAIL; }

	if(inlen != NULL)
	{
		ret = usb_bulk_read(versaloon_device_handle, versaloon_epin, (char *)versaloon_buf, versaloon_buf_size, versaloon_to);
		if(ret > 0)
		{
			*inlen = (uint16)ret;
			return ERROR_OK;
		}
		else
		{
			LOG_ERROR("usb receive failed with %d\n", ret);
			return ERROR_FAIL;
		}
	}
	else
	{
		return ERROR_OK;
	}
}

#define VERSALOON_RETRY_CNT				10
RESULT versaloon_init(void)
{
	uint16 ret = 0;
	uint8 retry;
	int verbosity_tmp, to_tmp;

	versaloon_device_handle = find_usb_device(versaloon_vid, versaloon_pid);
	if(NULL == versaloon_device_handle) { LOG_ERROR("no device found with vid=%04x, pid = %04x\n", versaloon_vid, versaloon_pid); return ERROR_FAIL; }

	// malloc temporary buffer
	versaloon_buf = (uint8 *)malloc(versaloon_buf_size);
	if(NULL == versaloon_buf) { LOG_ERROR("not enough memory\n"); return ERROR_FAIL; }

	// connect to versaloon
	verbosity_tmp = verbosity;
	to_tmp = versaloon_to;
	// not output error message when connectting
	verbosity = -1;
	// 500ms delay when connect
	versaloon_to = 100;
	for(retry = 0; retry < VERSALOON_RETRY_CNT; retry++)
	{
		versaloon_buf[0] = VERSALOON_GET_INFO;
		if((ERROR_OK == versaloon_send_command(1, &ret)) && (ret >= 3)) { break; }
	}
	verbosity = verbosity_tmp;
	versaloon_to = to_tmp;
	if(VERSALOON_RETRY_CNT == retry) { free(versaloon_buf); versaloon_buf = NULL; LOG_ERROR("communicate with versaloon fail\n"); return ERROR_FAIL; }

	versaloon_buf[ret] = 0;
	versaloon_buf_size = versaloon_buf[0] + (versaloon_buf[1] << 8);
	LOG_INFO("%s\n", versaloon_buf + 2);

	// free temporary buffer
	free(versaloon_buf);
	versaloon_buf = NULL;

	versaloon_buf = (uint8 *)malloc(versaloon_buf_size);
	if(NULL == versaloon_buf) { LOG_ERROR("not enough memory\n"); return ERROR_FAIL; }

	return ERROR_OK;
}

RESULT versaloon_fini(void)
{
	if(versaloon_device_handle != NULL)
	{
		usb_close(versaloon_device_handle);
		versaloon_device_handle = NULL;

		if(versaloon_buf != NULL)
		{
			free(versaloon_buf);
			versaloon_buf = NULL;
		}
	}

	return ERROR_OK;
}

const char* versaloon_get_hardware_name(uint8 idx)
{
	if((idx > 0) && (idx < 3))
	{
		return versaloon_hardwares[idx - 1];
	}
	else
	{
		return "unknown hardware";
	}
}

RESULT versaloon_get_hardware(uint8 *hardware)
{
	uint16 inlen;

#ifdef PARAM_CHECK
	if(NULL == versaloon_buf) { LOG_BUG("buffer not valid\n"); return ERROR_FAIL; }
	if(NULL == hardware) { LOG_BUG("parameter invalid\n"); return ERROR_FAIL; }
#endif

	versaloon_buf[0] = VERSALOON_GET_HARDWARE;

	if((ERROR_OK != versaloon_send_command(1, &inlen)) || (inlen != 1))
	{
		LOG_ERROR("communicate with versaloon fail\n"); 
		return ERROR_FAIL;
	}
	else
	{
		*hardware = versaloon_buf[0];
		LOG_DEBUG("versaloon hardware is %s\n", versaloon_get_hardware_name(*hardware));
		return ERROR_OK;
	}
}

RESULT versaloon_get_target_voltage(uint16 *voltage)
{
	uint16 inlen;

#ifdef PARAM_CHECK
	if(NULL == versaloon_buf) { LOG_BUG("buffer not valid\n"); return ERROR_FAIL; }
	if(NULL == voltage) { LOG_BUG("parameter invalid\n"); return ERROR_FAIL; }
#endif

	versaloon_buf[0] = VERSALOON_GET_TVCC;

	if((ERROR_OK != versaloon_send_command(1, &inlen)) || (inlen != 2))
	{
		LOG_ERROR("communicate with versaloon fail\n"); 
		return ERROR_FAIL;
	}
	else
	{
		*voltage = versaloon_buf[0] + (versaloon_buf[1] << 8);
		return ERROR_OK;
	}
}









// capability
static uint16 versaloon_get_pin_remap(uint16 mask)
{
	return mask;
}
static uint16 versaloon_get_pin_map(uint16 mask)
{
	return mask;
}
// NULL function
RESULT versaloon_null(void)
{
	return ERROR_OK;
}
// SPI
RESULT versaloon_spi_init(void)
{
	return usbtospi_init();
}
RESULT versaloon_spi_fini(void)
{
	return usbtospi_fini();
}
RESULT versaloon_spi_config(uint16 kHz, uint8 cpol, uint8 cpha, uint8 first_bit)
{
	return usbtospi_config(VERSALOON_SPI_PORT, kHz, cpol, cpha, first_bit);
}
RESULT versaloon_spi_io(uint8 *out, uint8 *in, uint16 len, uint16 inpos, uint16 inlen)
{
	return usbtospi_io(VERSALOON_SPI_PORT, out, in, len, inpos, inlen);
}
// GPIO
RESULT versaloon_gpio_init(void)
{
	return usbtogpio_init();
}
RESULT versaloon_gpio_fini(void)
{
	return usbtogpio_fini();
}
RESULT versaloon_gpio_config(uint16 mask, uint16 direction)
{
	return usbtogpio_config(VERSALOON_GPIO_PORT, versaloon_get_pin_remap(mask), direction > 0);
}
RESULT versaloon_gpio_config_8bit(uint8 mask, uint8 direction)
{
	return usbtogpio_config(VERSALOON_GPIO_PORT, versaloon_get_pin_remap(mask), direction > 0);
}
RESULT versaloon_gpio_in(uint16 mask, uint16 *value)
{
	uint16 value_tmp;
	RESULT ret;

	ret = usbtogpio_in(VERSALOON_GPIO_PORT, versaloon_get_pin_remap(mask), &value_tmp);
	*value = versaloon_get_pin_map(value_tmp);

	return ret;
}
RESULT versaloon_gpio_in_8bit(uint8 mask, uint8 *value)
{
	uint16 value_tmp;
	RESULT ret;

	ret = usbtogpio_in(VERSALOON_GPIO_PORT, versaloon_get_pin_remap(mask), &value_tmp);
	*value = (uint8)versaloon_get_pin_map(value_tmp);

	return ret;
}
RESULT versaloon_gpio_out(uint16 mask, uint16 value)
{
	return usbtogpio_out(VERSALOON_GPIO_PORT, versaloon_get_pin_remap(mask), versaloon_get_pin_remap(value));
}
RESULT versaloon_gpio_out_8bit(uint8 mask, uint8 value)
{
	return usbtogpio_out(VERSALOON_GPIO_PORT, versaloon_get_pin_remap(mask), versaloon_get_pin_remap(value));
}
// Delay
RESULT versaloon_delay_ms(uint16 ms)
{
	return usbtodelay_delay(ms | 0x8000);
}
RESULT versaloon_delay_us(uint16 us)
{
	return usbtodelay_delay(us & 0x7FFF);
}
// ISSP
RESULT versaloon_issp_init(void)
{
	if(ERROR_OK != usbtoissp_init()) { return ERROR_FAIL; }
	return usbtoissp_config(VERSALOON_ISSP_PORT);
}
RESULT versaloon_issp_fini(void)
{
	return usbtoissp_fini();
}
RESULT versaloon_issp_enter_program_mode(uint8 mode)
{
	return usbtoissp_enter_program_mode(VERSALOON_ISSP_PORT, mode);
}
RESULT versaloon_issp_leave_program_mode(uint8 mode)
{
	return usbtoissp_leave_program_mode(VERSALOON_ISSP_PORT, mode);
}
RESULT versaloon_issp_wait_and_poll(void)
{
	return usbtoissp_wait_and_poll(VERSALOON_ISSP_PORT);
}
RESULT versaloon_issp_vector(uint8 operate, uint8 addr, uint8 data, uint8 *buf)
{
	return usbtoissp_vector(VERSALOON_ISSP_PORT, operate, addr, data, buf);
}
// JTAG
RESULT versaloon_jtaghl_init(void)
{
	return usbtojtaghl_init();
}
RESULT versaloon_jtaghl_fini(void)
{
	return usbtojtaghl_fini();
}
RESULT versaloon_jtaghl_config(uint16 kHz, uint8 ub, uint8 ua, uint16 bb, uint16 ba)
{
	return usbtojtaghl_config(VERSALOON_JTAGHL_PORT, kHz, ub, ua, bb, ba);
}
RESULT versaloon_jtaghl_tms(uint8 *tms, uint8 len)
{
	return usbtojtaghl_tmsbyte(VERSALOON_JTAGHL_PORT, tms, len);
}
RESULT versaloon_jtaghl_ir(uint8 *ir, uint8 len, uint8 idle, uint8 want_ret)
{
	return usbtojtaghl_ir(VERSALOON_JTAGHL_PORT, ir, len, idle, want_ret);
}
RESULT versaloon_jtaghl_dr(uint8 *dr, uint16 len, uint8 idle, uint8 want_ret)
{
	return usbtojtaghl_dr(VERSALOON_JTAGHL_PORT, dr, len, idle, want_ret);
}
// Commit
RESULT versaloon_peripheral_commit(void)
{
	return usbtoxxx_execute_command();
}



RESULT versaloon_init_capability(void *p)
{
	((programmer_info_t *)p)->init = versaloon_init;
	((programmer_info_t *)p)->fini = versaloon_fini;

	((programmer_info_t *)p)->interfaces = (SPI | GPIO | ISSP | JTAG);

	// SPI
	((programmer_info_t *)p)->spi_init = versaloon_spi_init;
	((programmer_info_t *)p)->spi_fini = versaloon_spi_fini;
	((programmer_info_t *)p)->spi_config = versaloon_spi_config;
	((programmer_info_t *)p)->spi_io = versaloon_spi_io;

	// GPIO
	((programmer_info_t *)p)->gpio_init = versaloon_gpio_init;
	((programmer_info_t *)p)->gpio_fini = versaloon_gpio_fini;
	((programmer_info_t *)p)->gpio_config = versaloon_gpio_config;
	((programmer_info_t *)p)->gpio_in = versaloon_gpio_in;
	((programmer_info_t *)p)->gpio_out = versaloon_gpio_out;

	// Delay
	((programmer_info_t *)p)->delayms = versaloon_delay_ms;
	((programmer_info_t *)p)->delayus = versaloon_delay_us;

	((programmer_info_t *)p)->peripheral_commit= versaloon_peripheral_commit;

	// ISSP
	((programmer_info_t *)p)->issp_init = versaloon_issp_init;
	((programmer_info_t *)p)->issp_fini = versaloon_issp_fini;
	((programmer_info_t *)p)->issp_enter_program_mode = versaloon_issp_enter_program_mode;
	((programmer_info_t *)p)->issp_leave_program_mode = versaloon_issp_leave_program_mode;
	((programmer_info_t *)p)->issp_wait_and_poll = versaloon_issp_wait_and_poll;
	((programmer_info_t *)p)->issp_vector = versaloon_issp_vector;
	((programmer_info_t *)p)->issp_commit = versaloon_peripheral_commit;

	// Target voltage
	((programmer_info_t *)p)->get_target_voltage = versaloon_get_target_voltage;

	// JTAG
	((programmer_info_t *)p)->jtag_init = versaloon_jtaghl_init;
	((programmer_info_t *)p)->jtag_fini = versaloon_jtaghl_fini;
	((programmer_info_t *)p)->jtag_config= versaloon_jtaghl_config;
	((programmer_info_t *)p)->jtag_hl_tms = versaloon_jtaghl_tms;
	((programmer_info_t *)p)->jtag_hl_ir = versaloon_jtaghl_ir;
	((programmer_info_t *)p)->jtag_hl_dr = versaloon_jtaghl_dr;
	((programmer_info_t *)p)->jtag_aux_io_init = versaloon_gpio_init;
	((programmer_info_t *)p)->jtag_aux_io_fini = versaloon_gpio_fini;
	((programmer_info_t *)p)->jtag_aux_io_config = versaloon_gpio_config_8bit;
	((programmer_info_t *)p)->jtag_aux_io_out = versaloon_gpio_out_8bit;
	((programmer_info_t *)p)->jtag_aux_io_in = versaloon_gpio_in_8bit;
	((programmer_info_t *)p)->jtag_hl_delay_us = versaloon_delay_us;
	((programmer_info_t *)p)->jtag_hl_delay_ms = versaloon_delay_ms;
	((programmer_info_t *)p)->jtag_hl_commit = versaloon_peripheral_commit;

	return ERROR_OK;
}

