/**************************************************************************
 *  Copyright (C) 2008 by Simon Qian                                      *
 *  SimonQian@SimonQian.com                                               *
 *                                                                        *
 *  Project:    vsprog                                                    *
 *  File:       at89s5x.c                                                 *
 *  Author:     SimonQian                                                 *
 *  Versaion:   See changelog                                             *
 *  Purpose:    functions for at89s5x                                     *
 *  License:    See license                                               *
 *------------------------------------------------------------------------*
 *  Change Log:                                                           *
 *      YYYY-MM-DD:     What(by Who)                                      *
 *      2008-10-29:     created(by SimonQian)                             *
 **************************************************************************/
#include "app_cfg.h"
#include "at89s5x.h"

const program_area_map_t s5x_program_area_map[] = 
{
	{APPLICATION, 'f', 1},
	{EEPROM, 'e', 1},
	{USER_SIG, 'u', 1},
	{FUSE, 'f', 0},
	{LOCK, 'l', 0},
	{0, 0}
};

const s5x_param_t s5x_chips_param[] = {
//	chip_name,		signature,	program_mode,	flash_page_size,	flash_page_num,	ee_page_size,	ee_page_num,	usrsig_page_size,	fuse_size,	pe_out
	{"at89s51",		0x001E5106,	0x03,			256,				16,				0,				0,				0,					0,			0x69},
	{"at89s52",		0x001E5206,	0x03,			256,				32,				0,				0,				0,					0,			0x69},
	{"at89ls51",	0x001E6106,	0x03,			256,				16,				0,				0,				0,					0,			0x69},
	{"at89ls52",	0x001E6206,	0x03,			256,				32,				0,				0,				0,					0,			0x69},
	{"at89s2051",	0x001E23FF,	0x03,			32,					64,				0,				0,				32,					1,			0xFF},
	{"at89s4051",	0x001E43FF,	0x03,			32,					128,			0,				0,				32,					1,			0xFF},
	{"at89s8253",	0x001E7300,	0x03,			64,					192,			32,				64,				64,					1,			0xFF},
};
static s5x_param_t s5x_chip_param;

static uint8* s5x_flash_buffer = NULL;
static uint32 s5x_flash_size = 0;
static uint8* s5x_usrsig_buffer = NULL;
static uint32 s5x_usrsig_size = 0;

static uint16 s5x_chip_flash_offset = 0;
static uint8 s5x_program_mode = 0;

static uint16 s5x_flash_byte_delay_us = 500;
static uint16 s5x_usrsig_byte_delay_us = 500;
static uint16 s5x_isp_frequency = 560;
static uint8 s5x_lock = 0;
static uint8 s5x_fuse = 0;
static uint32 s5x_chip_id = 0;

void s5x_usage(void)
{
	printf("usage of %s:\n"
			"vsprog [COMMON_ARGUMENTS] (-o(e|r|w|v)[f|l]) [-m[b|p]] [-fFREQUENCY] [-lLOCK]\n"
			"  <-w WAIT> set Wait time between serial byte write, in us\n"
			"  <-o(e|r|w|v)[f|l]> Operations (Erase|Read|Write|Verify)[Flash|Lock]\n"
			"  <-m(b|p)> program Mode(Byte|Page)\n"
			"  <-FFREQUENCY> set Frequency, in KHz\n"
			"  <-lLOCK> define Lock(1..4)\n"
			"  <-fFUSE> define Fuse\n"
		, S5X_STRING);
}

void s5x_support(void)
{
	uint32 i;

	printf("Support list of %s:\n", S5X_STRING);
	for(i = 0; i < (sizeof(s5x_chips_param) / sizeof(s5x_chips_param[0])); i++)
	{
		printf("%s: id = 0x%06x, flash_size = %d, fuse_size = %d, usrsig_size = %d\n", 
				s5x_chips_param[i].chip_name, 
				s5x_chips_param[i].signature, 
				s5x_chips_param[i].flash_page_size * s5x_chips_param[i].flash_page_num, 
				s5x_chips_param[i].fuse_size, 
				s5x_chips_param[i].usrsig_page_size);
	}
}

void s5x_usage_and_exit(int exit_code)
{
	s5x_usage();
	exit(exit_code);
}

RESULT s5x_parse_argument(char cmd, const char **argu, uint32 argu_num)
{
	switch(cmd)
	{
	case 'h':
		s5x_usage();
		break;
	case 'S':
		s5x_support();
		break;
	case 'F':
		// set Frequency
		if(argu_num != 1) { LOG_ERROR("args of %c error\n", cmd); s5x_usage_and_exit(1); }
		s5x_isp_frequency = (uint16)strtoul(argu[0], NULL, 0);

		break;
	case 'w':
		// set Wait time
		if(argu_num != 1) { LOG_ERROR("args of %c error\n", cmd); s5x_usage_and_exit(1); }
		s5x_flash_byte_delay_us = (uint16)strtoul(argu[0], NULL, 0);
		if(s5x_flash_byte_delay_us & 0x8000) { LOG_ERROR("too much delay\n"); s5x_usage_and_exit(1); }

		break;
	case 'l':
		// define Lock
		if(argu_num != 1) { LOG_ERROR("args of %c error\n", cmd); s5x_usage_and_exit(1); }
		s5x_lock = (uint8)strtoul(argu[0], NULL, 0);
		if((s5x_lock < 1) || (s5x_lock > 4)) { LOG_ERROR("lock mode is from 1 to 4\n"); s5x_usage_and_exit(1); }
		s5x_lock--;

		break;
	case 'f':
		// define Fuse
		if(argu_num != 1) { LOG_ERROR("args of %c error\n", cmd); s5x_usage_and_exit(1); }
		s5x_fuse = (uint8)strtoul(argu[0], NULL, 0);
		if(s5x_fuse > 0x0F) { LOG_ERROR("fuse is 4 bit wide\n"); s5x_usage_and_exit(1); }

		break;
	case 'm':
		// program Mode
		if((argu_num != 1) || (strlen(argu[0]) != 1)) { LOG_ERROR("args of %c error\n", cmd); s5x_usage_and_exit(1); }
		switch(argu[0][0])
		{
		case 'b':
			// Byte mode
			s5x_program_mode &= ~S5X_PAGE_MODE;
			s5x_program_mode |= S5X_BYTE_MODE;
			break;
		case 'p':
			// Page mode
			s5x_program_mode &= ~S5X_BYTE_MODE;
			s5x_program_mode |= S5X_PAGE_MODE;
			break;
		default:
			LOG_ERROR("args of %c error\n", cmd);
			s5x_usage_and_exit(1);
			break;
		}

		break;
	default:
		return ERROR_FAIL;
		break;
	}

	return ERROR_OK;
}

RESULT s5x_probe_chip(char *chip_name)
{
	uint32 i;

	for(i = 0; i < (sizeof(s5x_chips_param) / sizeof(s5x_chips_param[0])); i++)
	{
		if(!strcmp(s5x_chips_param[i].chip_name, chip_name))
		{
			return ERROR_OK;
		}
	}

	return ERROR_FAIL;
}

RESULT s5x_prepare_buffer(program_info_t pi)
{
	if(pi.app != NULL)
	{
		memset(pi.app, S5X_FLASH_CHAR, pi.app_size);
		s5x_flash_buffer = pi.app;
	}
	else
	{
		return ERROR_FAIL;
	}

	return ERROR_OK;
}

RESULT s5x_write_buffer_from_file_callback(uint32 address, uint32 seg_addr, uint8* data, uint32 length)
{
	uint32 mem_addr = address & 0x0000FFFF;

	if(seg_addr != 0) { LOG_ERROR("no segment address is supported in s5x hex\n"); return ERROR_FAIL; }
#ifdef PARAM_CHECK
	if((length > 0) && (NULL == data)) { LOG_ERROR("parameter invalid\n"); return ERROR_FAIL; }
#endif

	// flash from 0x00000000
	switch(address >> 16)
	{
	case 0x0000:
		if(NULL == s5x_flash_buffer) { return ERROR_FAIL; }
		mem_addr += s5x_chip_flash_offset;
		if((mem_addr >= s5x_chip_param.flash_size) || 
			(length > s5x_chip_param.flash_size) || 
			((mem_addr + length) > s5x_chip_param.flash_size))
		{
			LOG_ERROR("flash memory invalid\n");
			return ERROR_FAIL;
		}
		memcpy(s5x_flash_buffer + mem_addr, data, length);
		s5x_flash_size += (uint16)length;
		break;
	default:
		LOG_ERROR("memory invalid\n");
		return ERROR_FAIL;
		break;
	}

	return ERROR_OK;
}

RESULT s5x_fini(void)
{
	s5x_flash_buffer = NULL;

	return ERROR_OK;
}

RESULT s5x_init(program_info_t *pi, const char *dir, programmer_info_t *prog)
{
	uint8 i;
	operation_t opt_tmp = {0};

	dir = dir;

	if(strcmp(pi->chip_type, S5X_STRING)) { LOG_BUG("%s: Why call me? I'm not %s\n", S5X_STRING, pi->chip_type); return ERROR_FAIL; }

	if(NULL == pi->chip_name)
	{
		// auto detect
		LOG_INFO("chip name not defined, try auto-detecting\n");
		opt_tmp.read_operations = CHIP_ID;
		s5x_chip_param.pe_out = 0x69;

		if(ERROR_OK != s5x_program(opt_tmp, *pi, prog)) { LOG_ERROR("%s auto-detect failed\n", pi->chip_type); return ERROR_FAIL; }
		LOG_INFO("chip_name not defined, auto-detect signature=0x%06x\n", s5x_chip_id);
		for(i = 0; i < (sizeof(s5x_chips_param) / sizeof(s5x_chips_param[0])); i++)
		{
			if(s5x_chip_id == s5x_chips_param[i].signature)
			{
				memcpy(&s5x_chip_param, s5x_chips_param + i, sizeof(s5x_chip_param));
				s5x_chip_param.flash_size = s5x_chip_param.flash_page_num * s5x_chip_param.flash_page_size;
				s5x_chip_param.ee_size = s5x_chip_param.ee_page_num * s5x_chip_param.ee_page_size;
				pi->app_size = s5x_chip_param.flash_size;

				LOG_INFO("%s found\n", s5x_chip_param.chip_name);
				pi->chip_name = (char *)s5x_chip_param.chip_name;

				return ERROR_OK;
			}
		}
		LOG_ERROR("%s auto-detect failed\n", pi->chip_type);
	}
	else
	{
		for(i = 0; i < (sizeof(s5x_chips_param) / sizeof(s5x_chips_param[0])); i++)
		{
			if(!strcmp(s5x_chips_param[i].chip_name, pi->chip_name))
			{
				memcpy(&s5x_chip_param, s5x_chips_param + i, sizeof(s5x_chip_param));
				s5x_chip_param.flash_size = s5x_chip_param.flash_page_num * s5x_chip_param.flash_page_size;
				s5x_chip_param.ee_size = s5x_chip_param.ee_page_num * s5x_chip_param.ee_page_size;
				pi->app_size = s5x_chip_param.flash_size;

				return ERROR_OK;
			}
		}
	}

	return ERROR_FAIL;
}

RESULT s5x_get_mass_product_data_size(operation_t operations, program_info_t pi, uint32 *size)
{
	operations = operations;
	pi = pi;
	size = size;

	return ERROR_FAIL;
}

RESULT s5x_prepare_mass_product_data(operation_t operations, program_info_t pi, uint8 *buff)
{
	operations = operations;
	pi = pi;
	buff = buff;

	return ERROR_FAIL;
}






static programmer_info_t *p = NULL;

RESULT s5x_program(operation_t operations, program_info_t pi, programmer_info_t *prog)
{
#define get_target_voltage(v)					p->get_target_voltage(v)

#define spi_init()								p->spi_init()
#define spi_fini()								p->spi_fini()
#define spi_conf(speed)							p->spi_config((speed), SPI_CPOL_LOW, SPI_CPHA_1EDGE, SPI_MSB_FIRST)
#define spi_io(out, outlen, in, inpos, inlen)	p->spi_io((out), (in), (outlen), (inpos), (inlen))
#define reset_init()							p->gpio_init()
#define reset_fini()							p->gpio_fini()
#define reset_output()							p->gpio_config(GPIO_SRST, 1)
#define reset_input()							p->gpio_config(GPIO_SRST, 0)
#define reset_set()								p->gpio_out(GPIO_SRST, GPIO_SRST)
#define reset_clr()								reset_input()
#define delay_ms(ms)							p->delayms((ms) | 0x8000)
#define delay_us(us)							p->delayus((us) & 0x7FFF)

#define commit()								p->peripheral_commit()


	union {
		unsigned char bytes[4]; // for sending one byte at a time, in reverse order
		unsigned int value; // for assigning
	} chip_sig;
	uint16 voltage;
	uint8 cmd_buf[4];
	uint8 tmp8;
	uint16 i, j, k, addr_start, addr_end;
	uint8 *page_buf = NULL;
	RESULT ret = ERROR_OK;

	pi = pi;
	p = prog;

#ifdef PARAM_CHECK
	if(NULL == p) { LOG_BUG("invalid programmer\n"); return ERROR_FAIL; }
	if(((operations.write_operations & APPLICATION) || (operations.verify_operations & APPLICATION)) && ((NULL == s5x_flash_buffer) || (0 == s5x_flash_size))) { LOG_ERROR("buffer not allocated\n"); return ERROR_FAIL; }
#endif

	// valid check
	if(((operations.write_operations & LOCK) || (operations.verify_operations & LOCK)) && (s5x_lock > 3)) { LOG_ERROR("lock undefined, use -l\n"); return ERROR_FAIL; }

	// get target voltage
	if(ERROR_OK != get_target_voltage(&voltage)) { return ERROR_FAIL; }
	LOG_DEBUG("target runs at %d.%dV\n", voltage / 1000, voltage % 1000);
	if(voltage < 2700)
	{
		LOG_WARNING("no power on target detected\n");
	}

	// check mode
	if((operations.write_operations & APPLICATION) || (operations.verify_operations & APPLICATION) || (operations.read_operations & APPLICATION))
	{
		switch(s5x_program_mode & S5X_MODE_MASK)
		{
		case 0:
			LOG_WARNING("program mode not defined, use page mode by default\n");
			s5x_program_mode = S5X_PAGE_MODE;
		case S5X_PAGE_MODE:
			if(!(s5x_chip_param.program_mode & S5X_PAGE_MODE)) { LOG_ERROR("page mode not supported by %s\n", s5x_chip_param.chip_name); return ERROR_FAIL; }
			break;
		case S5X_BYTE_MODE:
			if(!(s5x_chip_param.program_mode & S5X_BYTE_MODE)) { LOG_ERROR("byte mode not supported by %s\n", s5x_chip_param.chip_name); return ERROR_FAIL; }
			break;
		default:
			return ERROR_FAIL;
			break;
		}
	}

	// malloc page buffer
	if((operations.write_operations & APPLICATION) || (operations.read_operations & APPLICATION))
	{
		page_buf = malloc(s5x_chip_param.flash_page_size);
		if(NULL == page_buf) { LOG_ERROR("not enough memory\n"); return ERROR_FAIL; }
	}

	// here we go
	// init
	spi_init();
	reset_init();

	// enter program mode
	if(s5x_isp_frequency > 0)
	{
		// use s5x_isp_frequency
		spi_conf(s5x_isp_frequency);

		// toggle reset
		reset_output();
		reset_set();
		delay_ms(100);
		reset_input();
		delay_ms(30);
		reset_output();
		reset_set();
		delay_ms(10);

		// enter into program mode command
		cmd_buf[0] = 0xAC;
		cmd_buf[1] = 0x53;
		cmd_buf[2] = 0x00;
		cmd_buf[3] = 0x00;
		tmp8 = 0;
		// ret[3] should be 0x69
		spi_io(cmd_buf, 4, &tmp8, 3, 1);
		if((ERROR_OK != commit()) || ((s5x_chip_param.pe_out != 0xFF) && (s5x_chip_param.pe_out != tmp8))) { LOG_ERROR("fail to enter into program mode, try a slower speed\n"); ret = ERROR_FAIL; goto leave_program_mode; }
	}
	else
	{
		// auto check frequency
		LOG_ERROR("isp frequency can not be 0\n");
		return ERROR_FAIL;
	}

	// read chip_id
	cmd_buf[0] = 0x28;
	cmd_buf[1] = 0x00;
	cmd_buf[2] = 0x00;
	cmd_buf[3] = 0x00;
	spi_io(cmd_buf, 4, &chip_sig.bytes[2], 3, 1);
	cmd_buf[0] = 0x28;
	cmd_buf[1] = 0x01;
	cmd_buf[2] = 0x00;
	cmd_buf[3] = 0x00;
	spi_io(cmd_buf, 4, &chip_sig.bytes[1], 3, 1);
	cmd_buf[0] = 0x28;
	cmd_buf[1] = 0x02;
	cmd_buf[2] = 0x00;
	cmd_buf[3] = 0x00;
	spi_io(cmd_buf, 4, &chip_sig.bytes[0], 3, 1);
	if(ERROR_OK != commit()) { LOG_ERROR("fail to read signature\n"); ret = ERROR_FAIL; goto leave_program_mode; }
	LOG_INFO("signature read is 0x%06x\n", chip_sig.value & 0x00FFFFFF);
	s5x_chip_id = chip_sig.value & 0x00FFFFFF;
	if(!(operations.read_operations & CHIP_ID))
	{
		if(chip_sig.value != s5x_chip_param.signature) { LOG_WARNING("signature unmatch, read=0x%06x, want=0x%06x\n", s5x_chip_id, s5x_chip_param.signature); }
	}
	else
	{
		goto leave_program_mode;
	}

	if(operations.erase_operations > 0)
	{
		LOG_INFO("erasing chip\n");
		pgbar_init("erasing chip |", "|", 
					0, 
					1, 
					PROGRESS_STEP, '=');

		cmd_buf[0] = 0xAC;
		cmd_buf[1] = 0x80;
		cmd_buf[2] = 0x00;
		cmd_buf[3] = 0x00;
		spi_io(cmd_buf, 4, NULL, 0, 0);
		delay_ms(500);
		if(ERROR_OK != commit()) { pgbar_fini(); LOG_ERROR("fail to erase chip\n"); ret = ERROR_FAIL; goto leave_program_mode; }

		pgbar_update(1);
		pgbar_fini();
		LOG_INFO("chip erased\n");
	}

	if(operations.write_operations & APPLICATION)
	{
		// program
		LOG_INFO("programming flash\n");

		addr_start = s5x_chip_flash_offset;
		if(s5x_program_mode & S5X_PAGE_MODE)
		{
			addr_end = (addr_start + s5x_flash_size + s5x_chip_param.flash_page_size - 1) & ~(s5x_chip_param.flash_page_size - 1);
			addr_start = addr_start & ~(s5x_chip_param.flash_page_size - 1);
			pgbar_init("writing flash |", "|", 
						addr_start, 
						addr_end, 
						PROGRESS_STEP, '=');

			for(i = addr_start; i < addr_end; i += s5x_chip_param.flash_page_size)
			{
				cmd_buf[0] = 0x50;
				cmd_buf[1] = (i >> 8) & 0xFF;
				if(s5x_chip_param.flash_page_size == 256)
				{
					spi_io(cmd_buf, 2, NULL, 0, 0);
				}
				else
				{
					cmd_buf[2] = i & 0xFF;
					spi_io(cmd_buf, 3, NULL, 0, 0);
				}
				for(j = 0; j < s5x_chip_param.flash_page_size; j++)
				{
					spi_io(&s5x_flash_buffer[i + j], 1, NULL, 0, 0);
					delay_us(s5x_flash_byte_delay_us);
				}
				if(ERROR_OK != commit()) { pgbar_fini(); LOG_ERROR("fail to program flash in page mode\n"); ret = ERROR_FAIL; goto leave_program_mode; }

				pgbar_update(s5x_chip_param.flash_page_size);
			}
		}
		else if(s5x_program_mode & S5X_BYTE_MODE)
		{
			addr_end = addr_start + (uint16)s5x_flash_size;
			pgbar_init("writing flash |", "|", 
						addr_start, 
						addr_end, 
						PROGRESS_STEP, '=');

			j = 0;
			for(i = addr_start; i < addr_end; i++)
			{
				cmd_buf[0] = 0x40;
				cmd_buf[1] = (uint8)(i >> 8);
				cmd_buf[2] = (uint8)(i >> 0);
				cmd_buf[3] = s5x_flash_buffer[i];
				spi_io(cmd_buf, 4, NULL, 0, 0);
				delay_us(s5x_flash_byte_delay_us);

				j++;
				if((j >= s5x_chip_param.flash_page_size) || (i == (addr_end - 1)))
				{
					if(ERROR_OK != commit()) { pgbar_fini(); LOG_ERROR("fail to program flash in byte mode\n"); ret = ERROR_FAIL; goto leave_program_mode; }
					pgbar_update(j);
					j = 0;
				}
			}
		}

		pgbar_fini();
		LOG_INFO("flash programmed for %dbytes\n", s5x_flash_size);
	}

	if(operations.read_operations & APPLICATION)
	{
		if(operations.verify_operations & APPLICATION)
		{
			LOG_INFO("verifying flash\n");
		}
		else
		{
			s5x_flash_size = s5x_chip_param.flash_size;
			LOG_INFO("reading flash\n");
		}

		addr_start = s5x_chip_flash_offset;
		if(s5x_program_mode & S5X_PAGE_MODE)
		{
			addr_end = (addr_start + s5x_flash_size + s5x_chip_param.flash_page_size - 1) & ~(s5x_chip_param.flash_page_size - 1);
			addr_start = addr_start & ~(s5x_chip_param.flash_page_size - 1);
			pgbar_init("reading flash |", "|", 
						addr_start, 
						addr_end, 
						PROGRESS_STEP, '=');

			for(i = addr_start; i < addr_end; i += s5x_chip_param.flash_page_size)
			{
				cmd_buf[0] = 0x30;
				cmd_buf[1] = (i >> 8) & 0xFF;
				if(s5x_chip_param.flash_page_size == 256)
				{
					spi_io(cmd_buf, 2, NULL, 0, 0);
				}
				else
				{
					cmd_buf[2] = i & 0xFF;
					spi_io(cmd_buf, 3, NULL, 0, 0);
				}
				for(j = 0; j < s5x_chip_param.flash_page_size; j++)
				{
					page_buf[j] = 0;
				}
				spi_io(page_buf, s5x_chip_param.flash_page_size, page_buf, 0, s5x_chip_param.flash_page_size);
				if(ERROR_OK != commit()) { pgbar_fini(); LOG_ERROR("fail to read flash in page mode\n"); ret = ERROR_FAIL; goto leave_program_mode; }

				for(j = 0; j < s5x_chip_param.flash_page_size; j++)
				{
					if(operations.verify_operations & APPLICATION)
					{
						if(page_buf[j] != s5x_flash_buffer[i + j])
						{
							pgbar_fini();
							LOG_ERROR("flash verify failed at addr0x%04x, read=0x%02x, want=0x%02x\n", 
										i + j, page_buf[j], s5x_flash_buffer[i + j]);
							ret = ERROR_FAIL;
							goto leave_program_mode;
						}
					}
					else
					{
						
					}
				}

				pgbar_update(s5x_chip_param.flash_page_size);
			}
		}
		else if(s5x_program_mode & S5X_BYTE_MODE)
		{
			addr_end = addr_start + (uint16)s5x_flash_size;
			pgbar_init("reading flash |", "|", 
						addr_start, 
						addr_end, 
						PROGRESS_STEP, '=');

			j = 0;
			for(i = addr_start; i < addr_end; i++)
			{
				cmd_buf[0] = 0x20;
				cmd_buf[1] = (uint8)(i >> 8);
				cmd_buf[2] = (uint8)(i >> 0);
				cmd_buf[3] = 0;
				spi_io(cmd_buf, 4, page_buf + j, 3, 1);

				j++;
				if((j >= s5x_chip_param.flash_page_size) || (i == (addr_end - 1)))
				{
					if(ERROR_OK != commit()) { pgbar_fini(); LOG_ERROR("fail to read flash in byte mode\n"); ret = ERROR_FAIL; goto leave_program_mode; }
					pgbar_update(j);

					for(k = 0; k < j; k++)
					{
						if(operations.verify_operations & APPLICATION)
						{
							if(page_buf[k] != s5x_flash_buffer[i + 1 - j + k])
							{
								pgbar_fini();
								LOG_ERROR("flash verify failed at addr0x%04x, read=0x%02x, want=0x%02x\n", 
											i + 1 - j + k, page_buf[k], s5x_flash_buffer[i + 1 - j + k]);
								ret = ERROR_FAIL;
								goto leave_program_mode;
							}
						}
						else
						{
							
						}
					}
					j = 0;
				}
			}
		}

		pgbar_fini();
		if(operations.verify_operations & APPLICATION)
		{
			LOG_INFO("flash verified for %dbytes\n", s5x_flash_size);
		}
		else
		{
			LOG_INFO("flash read\n");
		}
	}

	if(operations.write_operations & USER_SIG)
	{
		if(0 == s5x_chip_param.usrsig_page_size)
		{
			LOG_ERROR("no usrsig supported on %s\n", s5x_chip_param.chip_name);
			ret = ERROR_FAIL;
			goto leave_program_mode;
		}

		LOG_INFO("programming usrsig\n");

		if(s5x_program_mode & S5X_PAGE_MODE)
		{
			pgbar_init("writing usrsig |", "|", 
						0, 
						1, 
						PROGRESS_STEP, '=');

			cmd_buf[0] = 0x52;
			cmd_buf[1] = 0x00;
			cmd_buf[2] = 0x00;
			spi_io(cmd_buf, 3, NULL, 0, 0);

			for(i = 0; i < s5x_chip_param.usrsig_page_size; i++)
			{
				spi_io(&s5x_usrsig_buffer[i], 1, NULL, 0, 0);
				delay_us(s5x_usrsig_byte_delay_us);
			}
			if(ERROR_OK != commit()) { pgbar_fini(); LOG_ERROR("fail to read usrsig in page mode\n"); ret = ERROR_FAIL; goto leave_program_mode; }

			pgbar_update(1);
		}
		else if(s5x_program_mode & S5X_BYTE_MODE)
		{
			pgbar_init("writing usrsig |", "|", 
						0, 
						1, 
						PROGRESS_STEP, '=');

			for(i = 0; i < s5x_chip_param.usrsig_page_size; i++)
			{
				cmd_buf[0] = 0x42;
				cmd_buf[1] = 0x00;
				cmd_buf[2] = (uint8)i;
				cmd_buf[3] = s5x_usrsig_buffer[i];
				spi_io(cmd_buf, 4, NULL, 0, 0);
				delay_us(s5x_usrsig_byte_delay_us);
			}
			if(ERROR_OK != commit()) { pgbar_fini(); LOG_ERROR("fail to program usrsig in byte mode\n"); ret = ERROR_FAIL; goto leave_program_mode; }
			pgbar_update(1);
		}

		pgbar_fini();
		LOG_INFO("usrsig programmed for %dbytes\n", s5x_usrsig_size);
	}

	if(operations.read_operations & USER_SIG)
	{
		if(0 == s5x_chip_param.usrsig_page_size)
		{
			LOG_ERROR("no usrsig supported on %s\n", s5x_chip_param.chip_name);
			ret = ERROR_FAIL;
			goto leave_program_mode;
		}

		if(operations.verify_operations & USER_SIG)
		{
			LOG_INFO("verifying usrsig\n");
		}
		else
		{
			LOG_INFO("reading usrsig\n");
		}

		if(s5x_program_mode & S5X_PAGE_MODE)
		{
			pgbar_init("reading usrsig |", "|", 
						0, 
						1, 
						PROGRESS_STEP, '=');

			cmd_buf[0] = 0x32;
			cmd_buf[1] = 0x00;
			cmd_buf[2] = 0x00;
			spi_io(cmd_buf, 3, NULL, 0, 0);

			for(i = 0; i < s5x_chip_param.usrsig_page_size; i++)
			{
				page_buf[i] = 0;
			}
			spi_io(page_buf, s5x_chip_param.usrsig_page_size, page_buf, 0, s5x_chip_param.usrsig_page_size);
			if(ERROR_OK != commit()) { pgbar_fini(); LOG_ERROR("fail to read usrsig in page mode\n"); ret = ERROR_FAIL; goto leave_program_mode; }

			for(i = 0; i < s5x_chip_param.usrsig_page_size; i++)
			{
				if(operations.verify_operations & USER_SIG)
				{
					if(page_buf[i] != s5x_usrsig_buffer[i])
					{
						pgbar_fini();
						LOG_ERROR("usrsig verify failed at addr0x%04x, read=0x%02x, want=0x%02x\n", 
									i, page_buf[i], s5x_usrsig_buffer[i]);
						ret = ERROR_FAIL;
						goto leave_program_mode;
					}
				}
				else
				{
					
				}
			}

			pgbar_update(1);
		}
		else if(s5x_program_mode & S5X_BYTE_MODE)
		{
			pgbar_init("reading usrsig |", "|", 
						0, 
						1, 
						PROGRESS_STEP, '=');

			for(i = 0; i < s5x_chip_param.usrsig_page_size; i++)
			{
				cmd_buf[0] = 0x22;
				cmd_buf[1] = 0x00;
				cmd_buf[2] = (uint8)i;
				cmd_buf[3] = 0;
				spi_io(cmd_buf, 4, page_buf + i, 3, 1);
			}

			if(ERROR_OK != commit()) { pgbar_fini(); LOG_ERROR("fail to read usrsig in byte mode\n"); ret = ERROR_FAIL; goto leave_program_mode; }
			for(i = 0; i < s5x_chip_param.usrsig_page_size; i++)
			{
				if(operations.verify_operations & USER_SIG)
				{
					if(page_buf[i] != s5x_usrsig_buffer[i])
					{
						pgbar_fini();
						LOG_ERROR("usrsig verify failed at addr0x%04x, read=0x%02x, want=0x%02x\n", 
									i, page_buf[i], s5x_usrsig_buffer[i]);
						ret = ERROR_FAIL;
						goto leave_program_mode;
					}
				}
				else
				{
					
				}
			}

			pgbar_update(1);
		}

		pgbar_fini();
		if(operations.verify_operations & USER_SIG)
		{
			LOG_INFO("usrsig verified for %dbytes\n", s5x_usrsig_size);
		}
		else
		{
			LOG_INFO("usrsig read\n");
		}
	}

	if(operations.write_operations & FUSE)
	{
		if(0 == s5x_chip_param.fuse_size)
		{
			LOG_ERROR("no fuse supported on %s\n", s5x_chip_param.chip_name);
			ret = ERROR_FAIL;
			goto leave_program_mode;
		}

		LOG_INFO("programming fuse\n");
		pgbar_init("writing fuse |", "|", 
					0, 
					1, 
					PROGRESS_STEP, '=');

		cmd_buf[0] = 0xAC;
		cmd_buf[1] = 0x10 + (s5x_fuse & 0x0F);
		cmd_buf[2] = 0x00;
		cmd_buf[3] = 0x00;
		spi_io(cmd_buf, 4, NULL, 0, 0);
		delay_ms(100);
		if(ERROR_OK != commit()) { pgbar_fini(); LOG_ERROR("fail to program fuse\n"); ret = ERROR_FAIL; goto leave_program_mode; }

		pgbar_update(1);
		pgbar_fini();
		LOG_INFO("fuse programmed\n");
	}

	if(operations.read_operations & FUSE)
	{
		if(0 == s5x_chip_param.fuse_size)
		{
			LOG_ERROR("no fuse supported on %s\n", s5x_chip_param.chip_name);
			ret = ERROR_FAIL;
			goto leave_program_mode;
		}
		if(operations.verify_operations & FUSE)
		{
			LOG_INFO("verifying fuse\n");
		}
		else
		{
			LOG_INFO("reading fuse\n");
		}
		pgbar_init("reading fuse |", "|", 
					0, 
					1, 
					PROGRESS_STEP, '=');

		cmd_buf[0] = 0x21;
		cmd_buf[1] = 0x00;
		cmd_buf[2] = 0x00;
		cmd_buf[3] = 0x00;
		spi_io(cmd_buf, 4, &tmp8, 3, 1);
		if(ERROR_OK != commit()) { pgbar_fini(); LOG_ERROR("fail to read lock bit\n"); ret = ERROR_FAIL; goto leave_program_mode; }

		pgbar_update(1);
		pgbar_fini();

		tmp8 &= 0x0F;
		if(operations.verify_operations & LOCK)
		{
			if(tmp8 == s5x_fuse)
			{
				LOG_INFO("fuse verified\n");
			}
			else
			{
				LOG_ERROR("fuse verify failed, read=%d, want=%d\n", tmp8, s5x_fuse);
				ret = ERROR_FAIL;
				goto leave_program_mode;
			}
		}
		else
		{
			LOG_INFO("fuse read is %d\n", tmp8);
		}
	}

	if(operations.write_operations & LOCK)
	{
		LOG_INFO("programming lock\n");
		pgbar_init("writing lock |", "|", 
					0, 
					1, 
					PROGRESS_STEP, '=');

		if(s5x_lock > 0)
		{
			for(i = 1; i < 4; i++)
			{
				if(s5x_lock >= i)
				{
					cmd_buf[0] = 0xAC;
					cmd_buf[1] = 0xE0 + (uint8)i;
					cmd_buf[2] = 0x00;
					cmd_buf[3] = 0x00;
					spi_io(cmd_buf, 4, NULL, 0, 0);
					delay_ms(100);
				}
			}
			if(ERROR_OK != commit()) { pgbar_fini(); LOG_ERROR("fail to program lock bit\n"); ret = ERROR_FAIL; goto leave_program_mode; }
		}
		else
		{
			LOG_INFO("no need to program lock mode 1\n");
		}

		pgbar_update(1);
		pgbar_fini();
		LOG_INFO("lock programmed\n");
	}

	if(operations.read_operations & LOCK)
	{
		uint8 lock;

		if(operations.verify_operations & LOCK)
		{
			LOG_INFO("verifying lock\n");
		}
		else
		{
			LOG_INFO("reading lock\n");
		}
		pgbar_init("reading lock |", "|", 
					0, 
					1, 
					PROGRESS_STEP, '=');

		cmd_buf[0] = 0x24;
		cmd_buf[1] = 0x00;
		cmd_buf[2] = 0x00;
		cmd_buf[3] = 0x00;
		spi_io(cmd_buf, 4, &tmp8, 3, 1);
		if(ERROR_OK != commit()) { pgbar_fini(); LOG_ERROR("fail to read lock bit\n"); ret = ERROR_FAIL; goto leave_program_mode; }

		lock = 0;
		for(i = 0; i < 3; i++)
		{
			if(tmp8 & (1 << (i + 2)))
			{
				lock += 1;
			}
		}

		pgbar_update(1);
		pgbar_fini();
		if(operations.verify_operations & LOCK)
		{
			if(s5x_lock == lock)
			{
				LOG_INFO("lock verified\n");
			}
			else
			{
				LOG_ERROR("lock verify failed, read=%d, want=%d\n", lock + 1, s5x_lock + 1);
				ret = ERROR_FAIL;
				goto leave_program_mode;
			}
		}
		else
		{
			LOG_INFO("lock read is %d\n", lock + 1);
		}
	}

leave_program_mode:
	// leave program mode
	reset_input();
	reset_fini();
	spi_fini();
	if(ERROR_OK != commit()) { LOG_ERROR("fail to exit program mode\n"); return ERROR_FAIL; }

	if(page_buf != NULL)
	{
		free(page_buf);
	}

	return ret;
}

