/**************************************************************************
 *  Copyright (C) 2008 by Simon Qian                                      *
 *  SimonQian@SimonQian.com                                               *
 *                                                                        *
 *  Project:    vsprog                                                    *
 *  File:       hex.c                                                     *
 *  Author:     SimonQian                                                 *
 *  Versaion:   See changelog                                             *
 *  Purpose:    hex file parser                                           *
 *  License:    See license                                               *
 *------------------------------------------------------------------------*
 *  Change Log:                                                           *
 *      YYYY-MM-DD:     What(by Who)                                      *
 *      2008-10-29:     created(by SimonQian)                             *
 **************************************************************************/
#include "app_cfg.h"

RESULT get_hex_from_str(uint8 *str, uint8 len, uint32 *value)
{
	char num[5], *ptr;

#ifdef PARAM_CHECK
	if((NULL == str) || (NULL == value) || (len > 4)) { LOG_BUG("parameter invalid\n\n"); return ERROR_FAIL; }
#endif

	memcpy(num, str, len);
	num[len] = 0;

	*value = strtoul(num, &ptr, 16);
	if((num + len) == ptr)
	{
		return ERROR_OK;
	}
	else
	{
		return ERROR_FAIL;
	}
}

RESULT write_hex_line_internal(FILE *hex_file, uint8 data_len, uint16 data_addr, uint8 type, uint8 *data)
{
	uint8 line_buf[10 + 0xFF * 2 + 2], checksum = 0, pos = 0, i;

#ifdef PARAM_CHECK
	if((NULL == hex_file) || ((data_len > 0) && (NULL == data))) { LOG_BUG("parameter invalid\n\n"); return ERROR_FAIL; }
#endif

	line_buf[pos] = ':';
	pos += 1;

	// data length
	checksum += data_len;
	sprintf((char *)line_buf + pos, "%02x", data_len);
	pos += 2;

	// address
	checksum += (uint8)data_addr;
	checksum += (uint8)(data_addr >> 8);
	sprintf((char *)line_buf + pos, "%04x", data_addr);
	pos += 4;

	// type
	checksum += type;
	sprintf((char *)line_buf + pos, "%02x", type);
	pos += 2;

	// data
	for(i = 0; i < data_len; i++)
	{
		checksum += data[i];
		sprintf((char *)line_buf + pos, "%02x", data[i]);
		pos += 2;
	}

	// checksum
	sprintf((char *)line_buf + pos, "%02x", checksum);
	pos += 2;

	// \n\r
	sprintf((char *)line_buf + pos, "\r\n");
	pos += 2;

	return (pos == fwrite(line_buf, 1, pos, hex_file));
}

RESULT write_hex_file_line(FILE *hex_file, uint32 data_addr, uint32 seg_addr, uint8 *data, uint8 len)
{
	static uint16 data_addr_ext_pre = 0;
	static uint32 seg_addr_pre = 0;

#ifdef PARAM_CHECK
	if((NULL == hex_file) || ((len > 0) && (NULL == data))) { return ERROR_FAIL; }
#endif

	if(0 == len)
	{
		// write end
		if(fwrite(":00000001ff\n\r", 1, 13, hex_file) != 13)
		{
			return ERROR_FAIL;
		}
		else
		{
			return ERROR_OK;
		}
	}

	// write addresses
	if(seg_addr_pre != seg_addr)
	{
		// write new seg_addr
		seg_addr_pre = seg_addr;
		if(ERROR_OK != write_hex_line_internal(hex_file, 4, (uint16)0x0000, HEX_TYPE_SEG_ADDR, (uint8 *)&seg_addr_pre)) { return ERROR_FAIL; }
	}
	if(data_addr_ext_pre != (data_addr >> 16))
	{
		// write new data_addr_ext
		data_addr_ext_pre = data_addr >> 16;
		if(ERROR_OK != write_hex_line_internal(hex_file, 2, (uint16)0x0000, HEX_TYPE_EXT_ADDR, (uint8 *)&data_addr_ext_pre)) { return ERROR_FAIL; }
	}

	// write data
	if(ERROR_OK != write_hex_line_internal(hex_file, len, (uint16)(data_addr & 0x0000FFFF), HEX_TYPE_DATA, data)) { return ERROR_FAIL; }

	return ERROR_OK;
}

RESULT read_hex_file(FILE *hex_file, HEX_WRITE_MEMORY_LINE_CALLBACK callback)
{
	uint8 line_buf[10 + 0xFF * 2 + 2], ch, pos, type, checksum;
	uint32 data_addr = 0, seg_addr = 0, length, tmp32, i;

#ifdef PARAM_CHECK
	if((NULL == hex_file) || (NULL == callback)) { return ERROR_FAIL; }
#endif

	rewind(hex_file);
	while(!feof(hex_file))
	{
		// ignore empty lines
		do{
			ch = (uint8)fgetc(hex_file);
		}while(('\n' == ch) || ('\r' == ch));
		if(ch != ':') { return ERROR_FAIL; }
		length = 0;
		// read line
		do{
			ch = (uint8)fgetc(hex_file);
			line_buf[length++] = ch;
		}while((ch != '\n') && (ch != '\r'));
		length -= 1;

		// process line
		if(length < 9) { return ERROR_FAIL; }
		checksum = 0;
		pos = 0;
		// data length
		if(ERROR_OK != get_hex_from_str(line_buf + pos, 2, &tmp32)) { return ERROR_FAIL; }
		pos += 2;
		checksum += (uint8)tmp32;
		if(length != (10 + tmp32 * 2)) { return ERROR_FAIL; }
		length = tmp32;
		// address
		if(ERROR_OK != get_hex_from_str(line_buf + pos, 4, &tmp32)) {return ERROR_FAIL; }
		pos += 4;
		checksum += (uint8)tmp32;
		checksum += (uint8)(tmp32 >> 8);
		data_addr = (data_addr & 0xFFFF0000) | (tmp32 & 0x0000FFFF);
		// type
		if(ERROR_OK != get_hex_from_str(line_buf + pos, 2, &tmp32)) { return ERROR_FAIL; }
		pos += 2;
		checksum += (uint8)tmp32;
		type = (uint8)tmp32;
		// data
		for(i = 0; i < length; i++)
		{
			if(ERROR_OK != get_hex_from_str(line_buf + pos, 2, &tmp32)) { return ERROR_FAIL; }
			pos += 2;
			checksum += (uint8)tmp32;
			line_buf[i] = (uint8)tmp32;
		}
		// checksum
		if(ERROR_OK != get_hex_from_str(line_buf + pos, 2, &tmp32)) { return ERROR_FAIL; }
		if(((checksum + tmp32) & 0xFF) != 0){ return ERROR_FAIL; }

		switch(type)
		{
		case HEX_TYPE_DATA:
			// data record
			if(ERROR_OK != callback(data_addr, seg_addr, line_buf, length)) { return ERROR_FAIL; }
			break;
		case HEX_TYPE_EOF:
			// end of file
			return ERROR_OK;
			break;
		case HEX_TYPE_SEG_ADDR:
			// segment address
			if(length > 4) { return ERROR_FAIL; }
			memcpy(&seg_addr, line_buf, length);
			break;
		case HEX_TYPE_EXT_ADDR:
			// extended address
			if(length > 2) { return ERROR_FAIL; }
			if(1 == length)
			{
				data_addr = (data_addr & 0x0000FFFF) | (line_buf[0] << 16);
			}
			else //if(2 == length)
			{
				data_addr = (data_addr & 0x0000FFFF) | (line_buf[0] << 24) | (line_buf[1] << 16);
			}
			break;
		default:
			LOG_WARNING("unknown type 0x%02x, current line ignored\n", type);
			break;
		}
	}

	return ERROR_FAIL;
}

